package jSimPack;

import jSimPack.SimTime.SimTimeUnit;
import java.util.*;
import java.util.concurrent.locks.ReentrantLock;

/**
 * Dispatches abstract events that occur at discrete points in time.  Events
 * are dispatched in increasing order of their release time.  The release time
 * of an event is calculated when it is scheduled via  {@link
 * #schedule(FutureEventListener,long,Object)}.  Each event is associated with
 * an <code>FutureEventListener</code> that takes an arbitrary action when the
 * event occurs.  While an <code>FutureEventListener</code> is processing an
 * event, it can freely add further events to, or remove other events from,
 * the event queue.  A <code>FutureEventList</code> keeps track of the current
 * time (as a <code>long</code>) in the simulation, so an event can never be
 * added with a timestamp in the "past" (before the current time), but events
 * can be removed before they are processed.  
 */
public class FutureEventList {

    private static long nextLabel = 0;
    /**
     * Holds the events.
     */
    private final Queue<FutureEvent> futureEvents;
    private final Queue<FutureEvent> systemEvents;
    /** Current simulation time */
    private SimTime wallClock;
    /** Time at which simulation should end */
    private SimTime endTime;
    /**
     * Whether the simulation should end immediately, regardless of
     * <code>endTime</code>.
     */
    private boolean endNow;
    private RandomSource randomSource;
    private double realtimeRate = 1.0;
    private boolean blockFlag = false;
    //breakpoint
    //private final Queue<SimTime> breakpoints;
    private final HashMap<SimTime, Breakpoint> breakpoints;
    private final ArrayList<BreakpointListener> breakpointListeners;

    /** used for stepping through the simulation */
    public FutureEventList(RandomSource randomSource) {
        wallClock = SimTime.ZERO;
        endNow = false;
        endTime = SimTime.FOREVER;
        futureEvents = new PriorityQueue<FutureEvent>();
        this.randomSource = randomSource;
        breakpoints = new HashMap<SimTime, Breakpoint>();
        systemEvents = new PriorityQueue<FutureEvent>();
        breakpointListeners = new ArrayList<BreakpointListener>();
    }

    public void setEndTime(SimTime time) {
        endTime = time;
    }

    public void setRealtimeRate(double rate) {
        if (rate < 0) {
            throw new IllegalArgumentException("negative rate: " + rate);
        }
        realtimeRate = rate;
    }

    public double getRealtimeRate() {
        return realtimeRate;
    }

    /**
     * Schedules the specified event to happen in the future.  The event will
     * happen <code>timeInterval</code> time ticks into the future.  The specified
     * <code>data</code> will be passed to {@link FutureEventListener#eventReleased(Object)}
     * when the event occurs.
     * 
     * @return
     * A unique object that can be passed to {@link #cancelEvent(Object, boolean)}
     * to identify this event.
     * 
     * @throws IllegalArgumentException
     * if <code>timeInterval</code> is negative
     */
    public FutureEvent schedule(FutureEventListener event, SimTime timeInterval, Object data) {
        if (timeInterval.isNegative()) {
            throw new IllegalArgumentException("negative timeInterval");
        }

        FutureEvent e = new FutureEvent(event,
                SimTime.add(wallClock, timeInterval), nextLabel++, data);
        futureEvents.offer(e);
        return e;
    }

    /**
     * Works just like scheduleEvent, but puts the event into the system queue (which are executed first and do not 
     * affect the random permutation of other events).
     */
    public FutureEvent scheduleNonsimulationEvent(FutureEventListener event, SimTime timeInterval, Object data) {
        if (timeInterval.isNegative()) {
            throw new IllegalArgumentException("negative timeInterval");
        }

        FutureEvent e = new FutureEvent(event,
                SimTime.add(wallClock, timeInterval), nextLabel++, data);
        systemEvents.offer(e);
        return e;
    }

    public boolean cancelNonsimulationEvent(FutureEvent event, boolean deleteNow) {
        if (!event.isScheduled()) {
            throw new IllegalArgumentException("event is not scheduled: " + event);
        }
        if (event.when.equals(wallClock)) {
            if (deleteNow) {
                event.expire();
            } else {
                return false;
            }
        } else {
            event.expire();
        }
        return true;
    }

    /**
     * Cancel the event associated with <code>label</code>. If the
     * event's release time coincides with the current simulation time,
     * the event will be removed if and only if <code>deleteNow</code>
     * is <code>true</code>.
     */
    public boolean cancelEvent(FutureEvent event, boolean deleteNow) {
        if (!event.isScheduled()) {
            throw new IllegalArgumentException("event is not scheduled: " + event);
        }
        if (event.when.equals(wallClock)) {
            if (deleteNow) {
                event.expire();
            } else {
                return false;
            }
        } else {
            event.expire();
        }
        return true;
    }

    public void runSimulationUntil(SimTime end) {
        endTime = end;
        runSimulation();
    }

    /**
     * Drives the simulation one step at a time until there are no more events
     * to be released, the simulation time exceeds the specified ending time, or
     * the simulation is ended immediately by calling {@link #endSimulation()}.
     */
    public void runSimulation() {
        while (!endNow && executeSimulationStep()) {
        // empty loop
        }
    }

    /**
     * Releases the next FutureEvent.
     * 
     * @return <code>true</code> if there are other events that have yet to be
     * released.
     */
    private boolean executeSimulationStep() {
        interleaveLock.lock();
        try {
            if (futureEvents.isEmpty()) {
                return false;
            }

            SimTime nextEventTime = futureEvents.peek().when;

            long nextEventNS = nextEventTime.getTruncNanoseconds();
            nextEventNS = nextEventNS - (Math.round((double) nextEventNS / 1000.0) * 1000);
            // if (nextEventNS != 0) {
            //     System.out.println("Noteworthy event:"  + nextEventTime);
            // }

            if (nextEventTime.isGreaterThan(endTime)) {
                return false;
            }

            SimTime nextSystemEventTime = systemEvents.peek().when;

            if (nextSystemEventTime.isLessThanOrEqual(nextEventTime)) {
                waitUntil(nextSystemEventTime);

                wallClock = nextSystemEventTime;

                while (!systemEvents.isEmpty() && systemEvents.peek().when.equals(nextSystemEventTime)) {
                    FutureEvent nextSystemEvent = systemEvents.remove();
                    if (nextSystemEvent.isScheduled()) {
                        nextSystemEvent.expire();
                        nextSystemEvent.handler.eventReleased(nextSystemEvent.callback);
                    }
                }

                return true;
            }

            //wait for realtime delay
            waitUntil(nextEventTime);

            wallClock = nextEventTime;

            //pull all the simultaneous events from the queue
            ArrayList<FutureEvent> eventBatch = new ArrayList<FutureEvent>();
            while (!futureEvents.isEmpty() && futureEvents.peek().when.equals(nextEventTime)) {
                eventBatch.add(futureEvents.remove());
            }

            //shuffle the events
            Collections.shuffle(eventBatch, randomSource.getRandom());

            for (FutureEvent e : eventBatch) {
                if (e.isScheduled()) {
                    e.expire();
                    e.handler.eventReleased(e.callback);
                }
            }
        } finally {
            interleaveLock.unlock();
        }
        
        return !futureEvents.isEmpty();
    }

    public void endSimulation() {
        endNow = true;
    }

    private void waitUntil(SimTime targetTime) {
        //compute the offest from start time in microseconds, multiply it by the realtime rate
        //the subtract from the target time to get the wait time
        //long start, end;
        //start = System.nanoTime();
        if (realtimeRate == Double.POSITIVE_INFINITY) {
            return;
        }
        if (realtimeRate == 0) {
            blockSimulation();
            return;
        }
        //System.out.println("waitUntil:targetTime  " + targetTime);
        //compute the amount of real time to wait based on the offset between the 
        long realtimeOffsetMs = (long) ((targetTime.getFracMilliseconds() - wallClock.getFracMilliseconds()) / realtimeRate);
        //System.out.println("realtimeOffsetMs = " + realtimeOffsetMs);
        if (realtimeOffsetMs > 0) {
            //use sleep to wait
            try {
                Thread.sleep(realtimeOffsetMs);
            } catch (InterruptedException ex) {
            //do nothing on interrupt
            }
        /*
        if (realtimeOffsetMs > 20) {
        try {
        Thread.sleep(realtimeOffsetMs); //convert wait time to milliseconds and sleep
        } catch (InterruptedException ex) {
        //do nothing on interrupt
        }
        } else {
        int i=0;
        while (new Date().getTime() - startTime.getTime() < 20) {
        i++;
        }
        System.out.println("Waitloop count = " + i);
        }*/
        }
    //end = System.nanoTime();
    //double elapsed = (end - start) / 1000000;
    //System.out.println("waitUntil:actualduration  " + elapsed + " ms");
    }

    private synchronized void blockSimulation() {
        blockFlag = true;
        while (blockFlag) {
            try {
                wait();
            } catch (InterruptedException ex) {
            //do nothing on interrupt
            }
        }
    }

    public synchronized void stepSimulation() {
        blockFlag = false;
        notifyAll();
    }

    /**
     * Add a breakpoint at the specified time.  The breakpoint stops execution
     * before any events that occur at that time.
     * 
     * @throws IllegalArgumentException if the specified time is in the past
     * or at the current simulation time
     *        
     * @param breakpointTime the time at which to pause execution
     *
     * @return <code>true</code> if the breakpoint was added,
     * <code>false</code> if there was already a breakpoint at that time.
     */
    public boolean addBreakpoint(SimTime breakpointTime) {
        if (breakpoints.containsKey(breakpointTime)) {
            return false;
        }
        breakpoints.put(breakpointTime, new Breakpoint(breakpointTime));
        return true;
    }

    public boolean removeBreakpoint(SimTime breakpointTime) {
        if (!breakpoints.containsKey(breakpointTime)) {
            return false;
        }
        //breakpoint exists so cancel the event and remove it from the list
        Breakpoint bp = breakpoints.remove(breakpointTime);
        cancelNonsimulationEvent(bp.breakpointEvent, true);
        return true;
    }

    private class Breakpoint implements FutureEventListener {

        public FutureEvent breakpointEvent;

        public Breakpoint(SimTime breakpointTime) {
            if (breakpointTime.isLessThanOrEqual(wallClock)) {
                throw new IllegalArgumentException("breakpoint time is not in the future: " + breakpointTime);
            }
            scheduleNonsimulationEvent(this, breakpointTime, null);
        }

        public void eventReleased(Object data) {
            //stop simulation
            setRealtimeRate(0.0);
            //notify the listeners
            for (BreakpointListener l : breakpointListeners) {
                l.breakpointOccured(wallClock);
            }
            breakpoints.remove(this);
        }
    }

    public void addBreakpointListener(BreakpointListener l) {
        breakpointListeners.add(l);
    }

    public boolean removeBreakpointListener(BreakpointListener l) {
        return breakpointListeners.remove(l);
    }

    public boolean isBlocked() {
        return blockFlag;
    }

    public SimTime getWallClock() {
        return wallClock;
    }

    private void log(Object... msg) {
        simulator.framework.Harness.log("FutureEventList", msg);
    }

    public String getEventList() {

        StringBuffer sb = new StringBuffer("Wall Clock: " + wallClock + "\n");
        if (futureEvents.size() > 0) {
            sb.append("Event List:\n");

            FutureEvent[] eventList = (FutureEvent[]) futureEvents.toArray();
            for (FutureEvent e : eventList) {
                sb.append(e.toString() + "\n");
            }
        } else {
            sb.append("No events on Future Event List\n");
        }
        return sb.toString();
    }

    public static void test1() {
        FutureEventList fel = new FutureEventList(new RandomSource());
        //fel.setRealtimeRate(Double.POSITIVE_INFINITY);
        for (int i = 0; i < 10; i++) {
            SimTime startTime = new SimTime(i, SimTime.SimTimeUnit.SECOND);
            PrintEvent p = new PrintEvent("Event " + i + " at " + startTime);
            Object label = fel.schedule(p, startTime, p);
            if (i == 3) {
            // fel.cancelEvent(label, true);
            }
        }
        fel.runSimulationUntil(new SimTime(20, SimTime.SimTimeUnit.SECOND));
    }

    public static void test2() {
        FutureEventList fel = new FutureEventList(new RandomSource());
        //fel.setRealtimeRate(Double.POSITIVE_INFINITY);
        for (int i = 0; i < 10; i++) {
            SimTime startTime = new SimTime(i, SimTime.SimTimeUnit.MILLISECOND);
            PrintEvent p = new PrintEvent("Event " + i + " at " + startTime);
            Object label = fel.schedule(p, startTime, p);
            if (i == 3) {
            // fel.cancelEvent(label, true);
            }
        }
        fel.runSimulationUntil(new SimTime(20, SimTime.SimTimeUnit.SECOND));
    }

    public static void test3() {
        final FutureEventList fel = new FutureEventList(new RandomSource());
        //fel.setRealtimeRate(Double.POSITIVE_INFINITY);
        double[] speeds = {
            0.1,
            0.5,
            1.0,
            2.0,
            10.0,
            100.0,
            1000.0,
            10000.0,
            Double.POSITIVE_INFINITY
        };
        int timeOffset = 1;
        for (Double speed : speeds) {
            SimTime startTime = new SimTime(timeOffset, SimTime.SimTimeUnit.SECOND);
            final double speed_ = speed;
            FutureEventListener speedSetter = new FutureEventListener() {

                public void eventReleased(Object data) {
                    System.out.println("Setting speed to " + speed_);
                    fel.setRealtimeRate(speed_);
                }
            };
            fel.schedule(speedSetter, startTime, null);
            timeOffset++;
        }
        SimTime startTime = new SimTime(timeOffset, SimTime.SimTimeUnit.SECOND);
        PrintEvent p = new PrintEvent("Final Event");
        fel.schedule(p, startTime, p);

        fel.runSimulationUntil(new SimTime(40, SimTime.SimTimeUnit.SECOND));
    }

    public static void breakpointTest() {
        final FutureEventList fel = new FutureEventList(new RandomSource());
        fel.setRealtimeRate(1.0);
        //register the print listener
        fel.addBreakpointListener(new BreakpointPrinter(System.out));
        //register a listener that resets the rate to 1.0
        fel.addBreakpointListener(new BreakpointListener() {

            public void breakpointOccured(SimTime breakpointTime) {
                System.out.println("Resetting runtime rate to 1.0 at time " + fel.getWallClock());
                fel.setRealtimeRate(1.0);
            }
        });

        //register some breakpoints
        fel.addBreakpoint(new SimTime(1, SimTimeUnit.SECOND));
        fel.addBreakpoint(new SimTime(2500, SimTimeUnit.MILLISECOND));
        fel.addBreakpoint(new SimTime(8500, SimTimeUnit.MILLISECOND));
        fel.addBreakpoint(new SimTime(8505, SimTimeUnit.MILLISECOND));
        //this breakpoint should not execute because it is after the last event
        fel.addBreakpoint(new SimTime(10, SimTimeUnit.SECOND));


        for (int i = 0; i < 10; i++) {
            SimTime startTime = new SimTime(i, SimTime.SimTimeUnit.SECOND);
            PrintEvent p = new PrintEvent("Event " + i + " at " + startTime);
            fel.schedule(p, startTime, p);
        }
        fel.runSimulationUntil(new SimTime(20, SimTime.SimTimeUnit.SECOND));
    }

    public static void sleepTest() {
        long start, end;
        double elapsed;
        for (int i = 0; i < 50; i += 5) {
            start = System.nanoTime();
            try {
                Thread.sleep(i);
            } catch (InterruptedException ex) {
            //do nothing
            }
            end = System.nanoTime();
            elapsed = (end - start) / 1000000;
            System.out.println("Requested " + i + " ms.  Actual = " + elapsed + " ms.");
        }

        for (int i = 0; i < 2000; i += 100) {
            start = System.nanoTime();
            try {
                Thread.sleep(0, i);
            } catch (InterruptedException ex) {
            //do nothing
            }
            end = System.nanoTime();
            elapsed = (end - start);
            System.out.println("Requested " + i + " ns.  Actual = " + elapsed + " ns.");
        }
    }

    /*public static void test2() {
    FutureEventList fel = new FutureEventList();
    for (int i = 0; i < 50; ++i) {
    long when = Math.round(Long.MAX_VALUE * Math.random());
    fel.schedule(new PrintEvent("" + when), when, null);
    }
    fel.runSimulation();
    }*/
    public static void main(String[] Args) {
        //test1();
        //System.out.println("\n****************************************\n");
        //test2();
        //test3();
        breakpointTest();
    //sleepTest();
    }

    private final ReentrantLock interleaveLock = new ReentrantLock(true);
    
    public void interleaveLock() {
        interleaveLock.lock();
    }

    public void interleaveUnlock() {
        interleaveLock.unlock();
    }
}
