/*
ECE649 Spring 2010
Group 7
Justin Ray/justinr2
(other names would go here)
 */
package simulator.elevatorcontrol;

// simulator.framework gives us access to CNI, PFI and Harness
import jSimPack.SimTime;
import simulator.elevatormodules.VendPositionCanPayloadTranslator;
import simulator.framework.*;
// simulator.payloads gives us access to all MessagePayload classes
import simulator.payloads.*;

/************************************
CLASS:
ButtonControl
AUTHOR:
Justin Ray
REVISION DATE:
CONSTRUCTOR PARAMETERS:
boolean verbose - whether or not to do verbose output
simTime period - the period the controller will run at
DESCRIPTION:
This controller class aligns the VC with the chute corresponding to the currently selected button
Source of:
 * mVendMotor (network)
 * VendMotor (framework)
Sink of:
 * mVendPosition[s] (network)
 * mButton[s] (network)
 * mCoinCount (network)
 * mVend (network)
 *************************************/
public class VendPositionControl extends Controller implements TimeSensitive {

    // define convenience state variables to track what state we are in
    public static enum State {

        IDLE,
        MOVE
    }
    // defines (in microseconds) how often this module should send out messages
    // and update its state
    SimTime period;
    // Physical and network interface objecst for ButtonControl
    //physical interface
    VendMotorPayload localVendMotor;
    //output network
    VendMotorCanPayloadTranslator mVendMotor;
    //input network
    VendPositionCanPayloadTranslator mVendPosition[];
    ButtonCanPayloadTranslator mButton[];
    CoinCountCanPayloadTranslator mCoinCount;
    VendCanPayloadTranslator mVend;
    //State variables
    State currentState;  // the current state machine state
    Direction appropriateDirection = Direction.STOP; //determines direction of movement
    //Constants

    public VendPositionControl(SimTime period, boolean verbose) {
        super("VendPositionControl", verbose);

        // set local variables to match constructor parameters
        this.period = period;

        //instantiate physical interface objects
        localVendMotor = new VendMotorPayload();
        physicalInterface.sendTimeTriggered(localVendMotor, period);

        // Register for fault messages
        //myFI.RegisterEventTriggered(MessagePayload.FaultEvent);

        //instantiate message objects
        CanMailbox p;
        /*
         * Note: we don't need to keep a reference to the actual mailbox (since we
         * only interact with them through the translators anyway), so we use
         * a local variable in the constructor for the mailbox.
         */

        p = new CanMailbox(MessageDictionary.VEND_CAN_ID);
        mVend = new VendCanPayloadTranslator(p);
        canInterface.registerTimeTriggered(p);

        p = new CanMailbox(MessageDictionary.COIN_COUNT_CAN_ID);
        mCoinCount = new CoinCountCanPayloadTranslator(p);
        canInterface.registerTimeTriggered(p);

        p = new CanMailbox(MessageDictionary.VEND_MOTOR_CAN_ID);
        mVendMotor = new VendMotorCanPayloadTranslator(p);
        canInterface.sendTimeTriggered(p, period);

        mVendPosition = new VendPositionCanPayloadTranslator[SodaMachine.NUM_CHUTES];
        mButton = new ButtonCanPayloadTranslator[SodaMachine.NUM_CHUTES];
        for (int i = 0; i < SodaMachine.NUM_CHUTES; i++) {
            int index = i + 1;

            p = new CanMailbox(MessageDictionary.VEND_POSITION_BASE_CAN_ID + ReplicationComputer.computeReplicationId(index));
            mVendPosition[i] = new VendPositionCanPayloadTranslator(p, index);
            canInterface.registerTimeTriggered(p);

            p = new CanMailbox(MessageDictionary.BUTTON_BASE_CAN_ID + ReplicationComputer.computeReplicationId(index));
            mButton[i] = new ButtonCanPayloadTranslator(p, index);
            canInterface.registerTimeTriggered(p);
        }

        // initialize the controller to the OFF state
        currentState = State.IDLE;

        //initialize the outputs to the IDLE state
        doIdle();

        // start a timer to interrupt based on the period
        timer.start(period);
    }

    // Necessary for extending the Networkable class
    public void timerExpired(Object callbackData) {

        //do some intermediate calculations

        //search for the currently selected button and position sensor
        int currentButton = SodaMachine.NONE;  //default value means no button pressed
        int currentPosition = SodaMachine.NONE;  //default value means we are not at any chute
        for (int i = 0; i < SodaMachine.NUM_CHUTES; i++) {
            int index = i + 1;
            if (mButton[i].getButtonState() == true) {
                currentButton = index;
            }
            if (mVendPosition[i].getValue() == true) {
                currentPosition = index;
            }
        }
        log("CurrentButton=",currentButton,"; CurrentPosition=",currentPosition);


        log("Executing " + currentState);
        //execute the state machine
        State newState = currentState; //if no transition occurs, remain in the same state
        switch (currentState) {
            case IDLE:
                //execute state actions
                doIdle();
                //check transitions
//#transition 'T5.1'
                if ((currentButton != SodaMachine.NONE && currentPosition != currentButton) &&
                        mCoinCount.getCoinCount() == SodaMachine.SODA_COST &&
                        mVend.getVendState() == false) {
                    log ("T5.1");
                    newState = State.MOVE;
//compute appropriate direction
                    if (currentButton < currentPosition) {
                        appropriateDirection = Direction.LEFT;
                    } else if (currentButton > currentPosition) {
                        appropriateDirection = Direction.RIGHT;
                    } else {
                        appropriateDirection = Direction.STOP;
                    }
                    log ("AppropriateDirection=" + appropriateDirection);
                }

                break;
            case MOVE:
                //execute state actions
                doMove(appropriateDirection);
                //check transitions
//#transition 'T5.2'
                if ((currentButton != SodaMachine.NONE && currentButton == currentPosition) ||
                        (currentPosition == 1 && localVendMotor.direction == Direction.LEFT) ||
                        (currentPosition == SodaMachine.NUM_CHUTES && localVendMotor.direction == Direction.RIGHT)) {
                    log ("T5.2");
                    newState = State.IDLE;

                }

                break;
            default:
                throw new RuntimeException("Unrecognized state " + currentState);
        }

        //advance to the next state we have computed
        if (currentState != newState) {
            log ("Tranisition: ",currentState," --> ",newState);
        }
        currentState = newState;

        // we start the timer to interrupt us again at the next period
        timer.start(period);
    }

    /**
     * actions for the IDLE state.
     */
    private void doIdle() {
        localVendMotor.set(Direction.STOP);
        mVendMotor.setDirection(Direction.STOP);
    }

    /**
     * actions for the EMPTY state.
     */
    private void doMove(Direction appropriateDirection) {
        localVendMotor.set(appropriateDirection);
        mVendMotor.setDirection(appropriateDirection);
    }
}
