package simulator.elevatormodules;

import jSimPack.SimTime;
import simulator.framework.*;
import simulator.payloads.*;

/**
 * Controls the actual vending of the soda from the chute into the tray.
 * Maintains the chute and tray state
 * 
 * @author Kenny Stauffer
 */
public class VendActuator extends Module {

    // for receiving weight additions/subtractions
    private VendActuatorPayload vendActuatorState;
    private VendPositionPayload[] vendPositionState;
    private SodaChutePayload[] sodaChutes;
    private SodaTrayPayload sodaTray;
    private VendMotorPayload vendMotorState;
    private int currentSodaType = 0;
    private boolean isVending = false;

    //private CarWeightPayload carWeightPhys;
    /**
     * Constructs a new <code>VendActuator</code> that sends messages every
     * <code>period</code> microseconds.
     * 
     * @param period
     * The time between <code>CarWeightPayload</code> and
     * <code>CarWeightAlarmPayload</code> messages.
     * 
     * @param verbose
     * Whether to generate log messages about passenger weights and the state of
     * the alarm when {@link #receive(CarWeightPayload)} is called.
     */
    public VendActuator(SimTime period, boolean verbose) {
        super(period, "VendActuator", verbose);

        vendActuatorState = new VendActuatorPayload();
        vendActuatorState.active = false;
        physicalConnection.registerEventTriggered(vendActuatorState);
        physicalConnection.registerTimeTriggered(vendActuatorState);
        //physicalConnection.sendTimeTriggered(vendActuatorState, period);
        //physicalConnection.sendOnce(vendActuatorState);

        sodaChutes = new SodaChutePayload[SodaMachine.NUM_CHUTES];
        vendPositionState = new VendPositionPayload[SodaMachine.NUM_CHUTES];

        for (int i=0; i < SodaMachine.NUM_CHUTES; i++) {
            int index = i+1;
            sodaChutes[i] = new SodaChutePayload(index);
            sodaChutes[i].set(SodaMachine.INITIAL_SODA_COUNT);
            physicalConnection.sendTimeTriggered(sodaChutes[i], period);
            physicalConnection.sendOnce(sodaChutes[i]);

            vendPositionState[i] = new VendPositionPayload(index);
            physicalConnection.registerEventTriggered(vendPositionState[i]);
        }

        sodaTray = new SodaTrayPayload();
        physicalConnection.sendTimeTriggered(sodaTray, period);
        physicalConnection.sendOnce(sodaTray);


        vendMotorState = new VendMotorPayload();
        physicalConnection.registerEventTriggered(vendMotorState);

    }

    @Override
    public void receive(VendPositionPayload mp) {
        currentSodaType = 0;
        for (int i=0; i < SodaMachine.NUM_CHUTES; i++) {
            if (vendPositionState[i].value == true) {
                currentSodaType = i+1;
            }
        }
    }

    @Override
    public void receive(VendActuatorPayload msg) {
        if (vendActuatorState.active && !isVending) {
            //first active message, so vend a soda into the tray
            isVending = true;
            if (currentSodaType == 0) {
                throw new RuntimeException("Error, attempted to vend a soda while not parked at a chute.");
            }
            if (vendMotorState.direction != Direction.STOP) {
                throw new RuntimeException("Error, attempted to vend a soda vend carriage is moving.");
            }
            sodaChutes[currentSodaType-1].sodaCount--;
            physicalConnection.sendOnce(sodaChutes[currentSodaType-1]);

            sodaTray.sodaValue = currentSodaType;
            physicalConnection.sendOnce(sodaTray);


        } else if (!vendActuatorState.active && isVending) {
            //first false message after vend
            isVending = false;
        }
    }

    @Override
    public void receive(VendMotorPayload msg) {
        //do nothing
    }



    public int getSodaFromTray() {
        int sodaType = sodaTray.getSodaValue();
        sodaTray.set(SodaMachine.NONE);
        physicalConnection.sendOnce(sodaTray);
        return sodaType;
    }


}
