package simulator.framework;

import jSimPack.SimTime;
import java.lang.ref.SoftReference;

/**
 * An event that is broadcast to components in the simulation.  Events are
 * dispatched by the Harness in order of their firing time.  An event can have
 * a message associated with it at construction time, and this message will be
 * delivered when the event is dispatched.  Components of the simulation
 * register to receive events by implementing {@link HarnessEventListener}
 * and calling {@link Harness#addListener(HarnessEventListener,int)}.
 *
 * These events are internal events relevant only to the simulation framework;
 * these are not used by the elevator controllers.
 *
 * When a HarnessEvent is created, it is assigned a unique sequence number
 * that can be used during debugging to follow the event through the
 * simulator.
 *
 * @author Christopher Martin
 * @author Kenny Stauffer
 */

public class HarnessEvent {

    enum State
    {
        VALID,
        CANCELED
    }

    private static long uniqueId = 1;

    private final long sequenceNumber;
    private final SimTime timeGenerated;
    private final SimTime delay;
    private final int type;
    private final HarnessEventListener source;
    private State state;
    private SoftReference<String> toString;

    /**
     * Creates a new event that will be dispatched after the specified length
     * of time.
     *
     * @param delay
     * The event will be dispatched this many simulated nanoseconds after the
     * current simulation time.
     * @param type
     * The type of this event.
     * @param source
     * The object that generated this event.
     */
    public HarnessEvent(SimTime delay, int type, HarnessEventListener source)
    {
        sequenceNumber = uniqueId++;
        timeGenerated = Harness.getTime();
        this.delay = delay;
        this.type = type;
        this.source = source;
        state = null;
    }

    public SimTime getDispatchDelay() {
        return delay;
    }

    public SimTime getGenerationTime() {
        return timeGenerated;
    }

    /**
     * Returns this events unique serial number.
     */
    public long getSequenceNumber() {
        return sequenceNumber;
    }

    public int getType() {
        return type;
    }

    public HarnessEventListener getSource() {
        return source;
    }

    void validate()
    {
        state = State.VALID;
    }

    void invalidate()
    {
        state = State.CANCELED;
    }

    boolean isValid()
    {
        return state == State.VALID;
    }

    @Override
    public String toString()
    {
        if (toString != null)
        {
            String result = toString.get();
            if(result!=null)
                return result;
        }
        StringBuilder sb = new StringBuilder("HarnessEvent(seqNum=");
        sb.append(sequenceNumber);
        sb.append(" timeGenerated=");
        sb.append(timeGenerated);
        sb.append(" delay=");
        sb.append(delay);
        sb.append(" type=");
        sb.append(type);
        sb.append(" source=");
        sb.append(source);
        sb.append(")");
        String result = sb.toString();
        toString = new SoftReference<String>(result);
        return result;
    }
}
