package simulator.framework;

import simulator.payloads.*;

/**
 * An adaptor for receiving event triggered messages.  An object that wishes
 * to receive time-triggered messages still needs to extend this class, but it
 * does not need to override any of these methods.  An object that wishes to
 * receive event-triggered messages, however, must extend this class and
 * override the appropriate <code>receive</code> method for each type of
 * message it wishes to receive.  There is no contract for the methods of this
 * class, so subclasses can take any implementation-specific actions.  The
 * implementations in this class throw an
 * <code>UnsupportedOperationException</code>, so if a subclass registers for
 * messages of a certain type but does not override the corresponding
 * <code>receive</code> method, a runtime exception will occur.
 *
 * @author Christopher Martin
 * @author Kenny Stauffer
 */
public abstract class Networkable {

    /**
     * A tag method to indicate that subclasses should override only specialized
     * methods.  Subclasses are supposed to provide methods tailored to the
     * types of messages they wish to receive.  Making this method
     * <tt>final</tt> enforces that contract.
     */
    protected final void receive(Payload msg) {
        throw new UnsupportedOperationException(
            "this method should never be called.  The method receive(" +
            msg.getClass() + ")" + " should be defined in the Networkable base class.");
    }

    public void receive(CanMailbox msg) {
        throw new UnsupportedOperationException(msg.toString());
    }

    public void receive(ButtonPayload msg) {
        throw new UnsupportedOperationException(msg.toString());
    }

    public void receive(CoinTrayPayload msg) {
        throw new UnsupportedOperationException(msg.toString());
    }

    public void receive(LightPayload msg) {
        throw new UnsupportedOperationException(msg.toString());
    }

    public void receive(VendMotorPayload msg) {
        throw new UnsupportedOperationException(msg.toString());
    }

    public void receive(SodaTrayPayload msg) {
        throw new UnsupportedOperationException(msg.toString());
    }

    public void receive(VendCarriagePositionPayload msg) {
        throw new UnsupportedOperationException(msg.toString());
    }

    public void receive(CoinReturnButtonPayload msg) {
        throw new UnsupportedOperationException(msg.toString());
    }

    public void receive(VendActuatorPayload msg) {
        throw new UnsupportedOperationException(msg.toString());
    }

    public void receive(VendPositionPayload msg) {
        throw new UnsupportedOperationException(msg.toString());
    }

    public void receive(CoinInPayload msg) {
        throw new UnsupportedOperationException(msg.toString());
    }

    public void receive(CoinOutPayload msg) {
        throw new UnsupportedOperationException(msg.toString());
    }

    public void receive(SodaChutePayload msg) {
        throw new UnsupportedOperationException(msg.toString());
    }

}
