package simulator.framework;

import jSimPack.SimTime;
import jSimPack.SimTime.SimTimeUnit;
import java.io.*;
import java.util.*;
import simulator.elevatormodules.passenger.Passenger;
import simulator.elevatormodules.passenger.PassengerControl;

public class PassengerInjector implements TimeSensitive {

    private boolean verbose;
    private final static SimTime INTER_PASSENGER_TIME = new SimTime(5000, SimTimeUnit.MILLISECOND);
    private final ArrayList<Passenger> finishedPassengers = new ArrayList<Passenger>();
    private final SystemTimer passengerStartTimer;
    private final int numPassengers;
    private final Random passengerRandom;
    private final PassengerControl pc;

    public PassengerInjector(int numPassengers, PassengerControl pc, boolean verbose) {
        this.numPassengers = numPassengers;
        this.pc = pc;
        this.verbose = verbose;
        passengerRandom = new Random(Harness.getRandomSeed());
        passengerStartTimer = new SystemTimer(this);

        startNewPassenger();
    }

    public void startNewPassenger() {
        Passenger p = new Passenger(this, pc, passengerRandom.nextLong(), verbose);
        passengerStartTimer.start(INTER_PASSENGER_TIME, p);
    }

    /**
     * Callback that gets called when the currently executing passenger is finished.
     */
    public void passengerFinished(Passenger p) {
        log("Finished Customer " + (finishedPassengers.size()+1));
        finishedPassengers.add(p);
        if (!allPassengersFinished()) {
            startNewPassenger();
        } else {
            Harness.endSim();
        }
    }

    public boolean allPassengersFinished() {
        return finishedPassengers.size() >= numPassengers;
    }

    public void timerExpired(Object data) {
        Passenger p = (Passenger)data;
        log("Injecting Customer " + (finishedPassengers.size()+1));
        p.start();
    }

    String getStats() {
        StringWriter output = new StringWriter();
        PrintWriter stats = new PrintWriter(output);

        int satisfiedCount= 0;
        int unsatistfiedCount = 0;

        stats.println("STATISTICS FOR CUSTOMERS");
        stats.println("------------------------");
        stats.println();

        for (Passenger p : finishedPassengers) {
            if (p.isSatisfied().isSatisfied) {
                satisfiedCount++;
            } else {
                unsatistfiedCount++;
            }
            stats.println(p.toString());
            stats.println();
        }

        if (numPassengers < finishedPassengers.size()) {

            stats.println();
            stats.println("WARNING:  " + numPassengers + " customers were requested, but only " + finishedPassengers.size() + " were injected.  You probably need to increase the runtime.");
            stats.println();
        }

        stats.println();
        stats.println("Customer Summary");
        stats.println("Satisfied: " + satisfiedCount);
        stats.println("Unsatisfied: " + unsatistfiedCount);
        stats.println("Total: " + finishedPassengers.size());

        stats.close();
        return output.toString();
    }
    
    String getSummaryStats() {
        StringWriter output = new StringWriter();
        PrintWriter stats = new PrintWriter(output);

        int satisfiedCount= 0;
        int unsatistfiedCount = 0;

        stats.println("STATISTICS FOR CUSTOMERS");
        stats.println("------------------------");
        stats.println();

        for (Passenger p : finishedPassengers) {
            if (p.isSatisfied().isSatisfied) {
                satisfiedCount++;
            } else {
                unsatistfiedCount++;
            }
        }

        if (numPassengers < finishedPassengers.size()) {

            stats.println();
            stats.println("WARNING:  " + numPassengers + " customers were requested, but only " + finishedPassengers.size() + " were injected.  You probably need to increase the runtime.");
            stats.println();
        }

        stats.println();
        stats.println("Customer Summary");
        stats.println("Satisfied: " + satisfiedCount);
        stats.println("Unsatisfied: " + unsatistfiedCount);
        stats.println("Total: " + finishedPassengers.size());

        stats.close();
        return output.toString();
    }

    private void log(String s) {
        if (verbose) {
            Harness.log("PassengerInjector", s);
        }
    }
}
