/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package simulator.framework;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;
import simulator.framework.CANNetwork.CanConnection;
import simulator.framework.NetworkScheduler.Connection;

import simulator.payloads.Payload;

/**
 * This class provides a basic framework for implementing runtime monitors.
 * This class is to be used for monitoring only.
 * No class that instantiates a descendent of RuntimeMonitor or contains a descendent
 * of Runtime Monitor may modify the system state in any way!  
 * That means that no controller may contain a RuntimeMonitor class.
 * 
 * If you wish to implement any time-triggered behavior that does not affect the
 * random aspects of the simulation, use the systemTimer object, which is a special
 * Timer that executes outside the simulation event queue.
 * 
 * @author Justin Ray
 */
public abstract class RuntimeMonitor extends Networkable implements TimeSensitive {

    final protected CanConnection canInterface = Harness.getCANNetwork().getCanConnection();
    final private Connection physicalInterface = Harness.getPhysicalNetwork().getFrameworkConnection(this);
    //data structures
    final protected SystemTimer systemTimer = new SystemTimer(this);
    private int warningCount = 0;
    private String name;

    /**
     * The monitor instantiates an object even for parts of the elevator that don't
     * exist (like calls and atFloors at non-existent landings) so you do not need
     * to worry about null values in the arrays and maps
     */
    public RuntimeMonitor() {
        this.name = "RuntimeMonitor";

//        //set up network inputs

    }

    public final int getWarningCount() {
        return warningCount;
    }

    public final String getWarningStats() {
        return name + " generated " + warningCount + " warnings.";
    }
    
    public final String getName() {
        return name;
    }
    
    private final void setName(String name) {
        this.name = name;
    }

    protected final void warning(String warning) {
        //System.err.format("[" + name + "] @%4.9f: WARNING:  %s", Harness.getTime().getFracSeconds(), warning);
        //System.err.println();
        Harness.log(name, "WARNING:  " + warning);
        warningCount++;
    }


    
    /**
     * Use reflection to create and return a monitor class of the specified name.
     * Note that the monitor must have a default (no arguments) constructor.
     * @param monitorName
     * @return
     */
    public final static RuntimeMonitor createMonitor(String monitorName) {
        ReflectionFactory rf = new ReflectionFactory();
        
        List<String> constructorArgs = new ArrayList<String>();
        List<String> packagePath = new ArrayList<String>();
        List<Object> parameters = new ArrayList<Object>();
        Payload payload = null;

        constructorArgs.add(monitorName);
        packagePath.add("simulator.elevatorcontrol.");
        packagePath.add("simulator.elevatormodules.");
        
        Constructor<?> cons = rf.prepareConstructor(packagePath,
            constructorArgs, parameters);
        RuntimeMonitor monitor = null;
        try {
            monitor = (RuntimeMonitor) cons.newInstance(parameters.toArray());
            monitor.setName(monitorName);
        } catch (InvocationTargetException ex) {
            throw new RuntimeException("Exception in RuntimeMonitor constructor: " + ex.getCause().toString(), ex);
        } catch (Exception e) {
            throw new RuntimeException("could not call constructor for RuntimeMonitor.", e);
        }
        return monitor;
    }
}
